/*

   QLTOOLS

   Read a QL floppy disk

   (c)1992 by Giuseppe Zanetti

   Giuseppe Zanetti
   via Vergani, 11 - 35031 Abano Terme (Padova) ITALY
   e-mail: beppe@sabrina.dei.unipd.it

   Valenti Omar
   via Statale,127 - 42013 Casalgrande (REGGIO EMILIA) ITALY
   e-mail: sinf7@imovx2.unimo.it
   sinf7@imoax1.unimo.it
   sinf@c220.unimo.it

   somewhat hacked by Richard Zidlicky, added formatting, -dl, -x option
   rdzidlic@cip.informatik.uni-erlangen.de

 */

#define TIME_DIFF    283996800

#define VERSION     "2.2,  Sep 1 1995"

/* Maximum allocation block (normally 3) */

#define MAXALB          6

/* Maximum number of sectors (norm. 1440) */

#define MAXSECT         2880

#if defined  (__NT__) || defined (__OS2__)
# define __MSDOS__
#endif

#if defined  (__MSDOS__)
# ifndef MSDOS
#  define MSDOS
# endif
#endif

#include <stdio.h>
#include <stdlib.h>
#ifndef MSDOS
#include <sys/time.h>
#include <unistd.h>		/* lseek read write */
#else
#include <time.h>
#endif
#include <fcntl.h>

#ifdef THINK_C
#include "console.h"
#include <string.h>
#else
#include <string.h>
#include <ctype.h>
#endif

#ifdef MSDOS
#include <bios.h>
#include <dos.h>		/* for delay    */
#include <io.h>			/* for setmode  */
#include <conio.h>

#define RESET   0
#define LAST    1
#define READ    2
#define WRITE   3
#define VERIFY  4
#define FORMAT  5
#endif

/* this include is neccessary since stdio.h in SUN 4.0 do not define it */

#ifndef SEEK_SET
#include <unistd.h>
#endif

int fd;

static unsigned char b[512 * MAXALB];	/* general purpose buffer */
static unsigned char b0[512 * MAXALB];
static unsigned char *map = b0 + 96;

static signed char *ltp;	/* ltp table */
static signed char *ptl;
static unsigned int bleod;	/* directory block offset */
static unsigned int byeod;	/*           bytes        */
static unsigned char *pdir;	/* memory image of directory */
static unsigned char *__base__;

static long err;
static short tranql = 1;
static short list = 0;

#ifdef MSDOS
static unsigned int drive;

#endif

/* globals */
static int gsides, gtracks, gsectors, goffset, allocblock, gclusters, gspcyl;
static int gssize = 512;
static int ql5a;		/* flag */
static int block_dir[64];
static int block_max = 0;
static long lac = 0;		/* last allocated cluster */

unsigned short wrd (unsigned char *wp);
void write_cluster (char *p, int num);
unsigned long int lwrd (unsigned char *p);
int read_cluster (char *p, int num);
void wr_wrd (unsigned char *wp, unsigned short r);
void usage (char *error);
void make_convtable (int verbose);
void wr_lwrd (unsigned char *p, unsigned long r);

void free_cluster (long);
int alloc_new_cluster (int, int);
void format (char *frmt, char *argfname);
void dir_write_back (void);
void set_header (int ni, long h);
void read_b0fat (int);
void write_b0fat (void);
void read_ltp (void);
void read_fat (void);
void read_dir (void);
void print_map (void);

#ifdef QDOS
#define SECT0 1
#else
#define SECT0 0
#endif

/* logical to physical translation macros */

#define LTP_TRACK(_sect_)   ((_sect_)/gspcyl)
#define LTP_SIDE(_sect_)    (0>ltp[(_sect_)%gspcyl])
#define LTP_SCT(_sect_) \
     (((0x7f&ltp[(_sect_)%gspcyl])+goffset*LTP_TRACK(_sect_)) % gsectors)
#define LTP_SECT(_sect_)    (LTP_SCT(_sect_)+1)

#ifdef QDOS
#define LTP(_sect_)  ((LTP_SIDE(_sect_)<<8)+(LTP_TRACK(_sect_)<<16)+ \
                      LTP_SCT(_sect_)+1)
#else
#define LTP(_sect_)   (gssize*(LTP_TRACK(_sect_)*gspcyl+ \
                       LTP_SIDE(_sect_)*gsectors+LTP_SCT(_sect_)))
#endif


/* some defines to hide FAT bit artistics, heavy use */
/* of global variables RZ */

#define FAT_FILE(__cluster) (__base__=map+(__cluster)*3,\
                          (unsigned long)((*__base__) <<4)+ \
                          (unsigned long)((*(__base__+1)>>4)&15))


#define FAT_CL(__cluster)  (__base__= map+(__cluster)*3,\
                          (*(__base__+1) &15)*256+ *(__base__+2))



#define SET_FAT_FILE(__cluster,__file)\
    (__base__=map+(__cluster)*3,\
     *(__base__)=(__file)>>4,\
     *(__base__+1)=(((__file)&15)<<4)+(*(__base__+1)&15))


#define SET_FAT_CL(__cluster,__clnum)\
    (__base__= map+(__cluster)*3,\
     *(__base__+1)=((__clnum)>>8)+(*(__base__+1)&(~15)),\
     *(__base__+2)=(__clnum)&255)

/* some defines for block0 info */

#define  SET_TRACKS(x) wr_wrd(b0+30,x)
#define  SET_SPT(x)    wr_wrd(b0+26,x)
#define  SET_SPCYL(x)  wr_wrd(b0+28,x)
#define  SET_SOFF(x)   wr_wrd(b0+38,x)
#define  SET_DIRBL(x)  wr_wrd(b0+34,x)
#define  SET_DIROF(x)  wr_wrd(b0+36,x)
#define  SET_FREE(x)   wr_wrd(b0+20,x)
#define  SET_TOTAL(x)  wr_wrd(b0+24,x)
#define  SET_GOOD(x)   wr_wrd(b0+22,x)
#define  SET_SPCL(x)   wr_wrd(b0+32,x)


#define  GET_TRACKS() wrd(b0+30)
#define  GET_SPT()    wrd(b0+26)
#define  GET_SPCYL()  wrd(b0+28)
#define  GET_SOFF()   wrd(b0+38)
#define  GET_DIRBL()  wrd(b0+34)
#define  GET_DIROF()  wrd(b0+36)
#define  GET_FREE()   wrd(b0+20)
#define  GET_TOTAL()  wrd(b0+24)
#define  GET_GOOD()   wrd(b0+22)
#define  GET_SPCL()   wrd(b0+32)

#ifndef min
# define min(a,b) (a<b ? a : b)
#endif

/* byte order conversion */

unsigned short wrd (unsigned char *wp)
{
    unsigned short r;

    r = (((unsigned short) *(wp)) << 8) + (unsigned short) *(wp + 1);
    return r;
}

void wr_wrd (unsigned char *wp, unsigned short r)
{
    *(wp) = r >> 8;
    *(wp + 1) = r & 255;

}
void wr_lwrd (unsigned char *p, unsigned long r)
{
    *(p + 3) = r & 255;
    r = r >> 8;
    *(p + 2) = r & 255;
    r = r >> 8;
    *(p + 1) = r & 255;
    *p = r >> 8;
}


unsigned long int lwrd (unsigned char *p)
{
    /* 16 bit PC compilers are _very_ fussy */
    return (long) wrd (p + 2) | (long) ((long) wrd (p) << 16);
}


void cat_file (long fnum)
{
    long int flen, file, blk;
    int i, ii, s, start, end;
    char buffer[512 * MAXALB], *entry;
    long qldata = 0;

    int ok, del;

#ifdef MSDOS
    setmode (fileno (stdout), O_BINARY);
#endif

    del = 0;
    if ((fnum << 6) >= byeod + bleod * gssize)
    {
	fprintf (stderr, "file number out of bounds\n");
	exit (1);
    }

    entry = pdir + (fnum << 6);
    flen = lwrd (entry);	/* with the header */
    if (*(entry + 5) == 1)
    {
	qldata = *(long *) (entry + 6);
    }

    if (flen + wrd (entry + 14) == 0)
    {
	fprintf (stderr, "warning: file %ld appears to be deleted, trying to recover it\n", fnum);
	fnum = (fnum & 0xf) + 0xfd0;
	del = 1;
    }

    s = 0;
    do
    {
	ok = 0;
	for (i = 0; i < gclusters; i++)
	{
	    file = FAT_FILE (i);
	    blk = FAT_CL (i);
	    if (file == fnum && blk == s)
	    {
		if (blk == s)
		{
		    read_cluster (buffer, i);
		    if (s == 0)
			if (flen == 0)
			    flen = lwrd (buffer);
#if 0				/* sandy controler doesn't maintain the header copy */
			else if (flen != lwrd (buffer))
			    fprintf (stderr, "read file: incomplete header copy \n");
#endif
		    if (s == 0)
			start = 64;
		    else
			start = 0;
		    end = gssize * allocblock;
		    if (s == (flen / (gssize * allocblock)))
			end = flen % (gssize * allocblock);
		    err = write (1, buffer + start, end - start);
		    if (err < 0)
			perror ("output file: write(): ");
		    ok = 1;
		}
		if (ok)
		    break;
	    }
	}

	if (!ok)
	{
	    fprintf (stderr, "\n\nCluster #%d of file #%ld not found\n\n", s, fnum);
	    if (s == 0 && del)
	    {
		fprintf (stderr, "sorry, not recoverable, block #0 missing\n\n");
#ifdef MSDOS
		setmode (1, O_BINARY);
#endif
		exit (1);
	    }
	    err = lseek (1, gssize * allocblock, SEEK_CUR);	/* leave hole */
	    if (err < 0)	/* non seekable */
		for (ii = 0; ii < allocblock * gssize / 16; i++)
		    write (1, "################", 16);
	}
	s++;
    }
    while (s <= flen / (gssize * allocblock));
    if (qldata)
    {
	long stuff[2];

	if (lseek (1, -8, SEEK_END) > 0)
	{
	    read (1, stuff, 8);
	    if (*stuff != *(long *) "XTcc")
	    {
		lseek (1, 0, SEEK_END);
		*stuff = *(long *) "XTcc";
		*(stuff + 1) = qldata;
		write (1, stuff, 8);
	    }
	}
    }
}

void del_file (long fnum)
{
    long int flen, file, blk;
    int i, freed, blk0;
    unsigned char *entry;

#ifdef MSDOS
    setmode (fileno (stdout), O_BINARY);
#endif

    freed = 0;
    blk0 = -1;
    entry = pdir + (fnum << 6);

    if ((fnum << 6) >= bleod * gssize + byeod)
    {
	fprintf (stderr, "file number out of bounds\n");
	exit (1);
    }

    flen = lwrd (entry);	/* with the header */
    if (flen == 0)
    {
	fprintf (stderr, "file already deleted?\n");
	exit (1);
    }

    for (i = 1; i < gclusters; i++)
    {
	file = FAT_FILE (i);
	if (file == fnum)
	{
	    if (FAT_CL (i) == 0)
		blk0 = i;
	    free_cluster (i);
	    freed++;
	}

#ifdef MSDOS
	setmode (fileno (stdout), O_BINARY);
#endif

    }

    blk = GET_FREE ();
    SET_FREE (freed * allocblock + blk);

    if (blk0 > 0)
    {
	read_cluster (b, blk0);
	memcpy (b, entry, 64);
	write_cluster (b, blk0);
    }
    else
	fprintf (stderr, "block 0 of file missing??\n");

    wr_wrd (entry + 14, 0);	/* lunghezza del nome del file */
    wr_lwrd (entry, 0);		/* lunghezza del file */
    *(entry + 5) = 0;

    write_b0fat ();		/* write_cluster(b0,0); */
    dir_write_back ();
}


void usage (char *error)
{
    fprintf (stderr, "error %s\n\n", error);

#ifdef MSDOS
    fprintf (stderr, "Usage: qltools [a:|b:] -[options] [filename]\n\n");
#else
#ifdef QDOS
    fprintf (stderr, "Usage: qltools drive -[options] [filename]\n\n");
#endif
#endif

    fprintf (stderr, "options:\n\n");
    fprintf (stderr, "-d         list directory\n");
    fprintf (stderr, "-dl        list deleted files as well\n");
    fprintf (stderr, "-s         list short directory\n");
    fprintf (stderr, "-i         list info\n");
    fprintf (stderr, "-m         list disk map\n");
    fprintf (stderr, "-c         list conversion table of sectors\n");
    fprintf (stderr, "-w <name>  write <name> file to disk\n");
    fprintf (stderr, "-xFN <size> make FN executable with dataspace=size\n");
    fprintf (stderr, "-rFN       remove file number FN\n");
    fprintf (stderr, "-nFN       copy file number FN to stdout, handles\n");
    fprintf (stderr, "               deleted files where possible\n");
    fprintf (stderr, "-fxx <name> format as xx=hd|dd disk with label <name>\n\n");

#ifdef MSDOS
    fprintf (stderr, "QLTOOLS for MS-DOS (version %s)\n\n");
#else
#ifdef QDOS
    fprintf (stderr, "drive is in the form flp1_*D2d \n");
    fprintf (stderr, "QLTOOLS for QDOS (version: %s)\n\n", VERSION);
#else
    fprintf (stderr, "diskimage is either a file with the image of a QL format disk\n");
    fprintf (stderr, "or a Unix device with a QL disk inserted in it (/dev/fd...)\n\n");
    fprintf (stderr, "QLTOOLS for UNIX (version: %s)\n\n", VERSION);
#endif
#endif

    fprintf (stderr, "Giuseppe Zanetti  ");
    fprintf (stderr, "via Vergani, 11 - 35031 Abano Terme (Padova) ITALY\n");
    fprintf (stderr, "e-mail: beppe@sabrina.dei.unipd.it\n");
    fprintf (stderr, "Version 2.01:Valenti Omar, Via Statale 127, Casalgrande (RE)\n");
    fprintf (stderr, "e-mail: sinf7@imovx2.unimo.it   sinf7@imoax1.unimo.it\n");
    fprintf (stderr, "Version 2.02:Richard Zidlicky, rdzidlic@cip.informatik.uni-erlangen.de\n");
    fprintf (stderr, "version 2.1: Jonathan Hudson, +968 699407 (voice/fax/data)\n");

    exit (1);
}

void print_info (void)
{
    int i;

    printf ("Disk ID          : ");

    for (i = 0; i < 4; i++)
    {
	printf ("%c", b0[i]);
    }
    printf ("\nDisk Label       : ");

    for (i = 0; i < 10; i++)
    {
	printf ("%c", b0[4 + i]);
    }


    printf ("\nsectors per track: %i\n", gsectors);
    printf ("sectors per cyl. : %i\n", gspcyl);
    printf ("number of cylind.: %i\n", gtracks);
    printf ("allocation block : %i\n", allocblock);
    printf ("sector offset/cyl: %i\n", goffset);
    printf ("random           : %04x\n", wrd (b0 + 0xe));
    printf ("Updates          : %ld\n", lwrd (b0 + 0x10));

    printf ("free sectors     : %i\n", GET_FREE ());
    printf ("good sectors     : %i\n", GET_GOOD ());
    printf ("total sectors    : %i\n", GET_TOTAL ());

    printf ("directory is     : %u sectors and %u bytes\n", bleod, byeod);

    printf ("\nlogical-to-physical sector mapping table:\n\n");
    for (i = 0; i < gspcyl; i++)
	printf ("%x ", (unsigned char) ltp[i]);
    printf ("\n");

    if (ql5a)
    {
	printf ("\nphysical-to-logical sector mapping table:\n\n");
	for (i = 0; i < gspcyl; i++)
	    printf ("%x ", (unsigned char) ptl[i]);
    }
    printf ("\n");
}

void print_entry (unsigned char *entry, int fnum, int del, int shortflag)
{
    unsigned char c;
    int i;

    if (entry == NULL)
	return;

    if (lwrd (entry) + wrd (entry + 14) == 0)
	return;

    if (!shortflag)
	printf ("%3i: ", fnum);

    printf (del ? "deleted  " : "         ");

    {
	short j;

	j = min (wrd (entry + 14), 36);

	for (i = 0; i < j; i++)
	{
	    c = *(entry + 16 + i);
	    printf ("%c", (c));
	}
    }

    if (!shortflag)
    {
	for (i = 0; i < 38 - wrd (entry + 14); i++)
	    printf (" ");

	switch (*(entry + 5))
	{
	    case 0:
		printf ("   \t\t");
		break;
	    case 1:
		printf ("E %8ld\t", lwrd (entry + 6));
		break;
	    case 2:
		printf ("Rel  \t\t");
		break;
	    case 3:
		printf ("Thor dir \t");
		break;
	    case 4:
		printf ("CST  dir \t");
		break;
	    case 255:
		printf ("->  \t\t");
		break;
	    default:
		printf ("(type %3d)", *(entry + 5));
		break;
	}

	printf ("    %ld", (long) (lwrd (entry) - 64L));
    }
}

void print_dir (int shortflag, int lflag)
{
    int d, i, del = 0, fn;
    unsigned char *entry;

    if (!shortflag)
    {
	for (i = 0; i < 10; i++)
	{
	    printf ("%c", b0[4 + i]);
	}
	printf ("\n");

	printf ("%i/%i sectors.\n\n", GET_FREE (), GET_GOOD ());
/*          printf("bleod %d\tbyeod %d\n",bleod,byeod); */
    }

    for (d = 64; d < gssize * bleod + byeod; d += 64)
    {
	entry = NULL;
	if (lwrd (pdir + d) + wrd (pdir + d + 14) == 0)
	{
	    if (lflag)
	    {
		for (i = 1; i < gclusters && entry == NULL; i++)
		{
		    fn = FAT_FILE (i);
		    if (FAT_CL (i) == 0 &&
			((fn & 0xff0) == 0xfd0) &&
			((fn & 0xf) == ((d >> 6) & 15)))
		    {
			read_cluster (b, i);
			entry = b;
			del = 1;
		    }
		}
	    }
	}
	else
	{
	    entry = pdir + d;
	    del = 0;
	}
	if (entry == NULL)
	    continue;

	print_entry (entry, d >> 6, del, shortflag);
	printf ("\n");
    }
}

void make_convtable (int verbose)
{
    int i, si, tr, se, uxs, sectors;

    if (verbose)
    {
	printf ("\nCONVERSION TABLE\n\n");
	printf ("logic\ttrack\tside\tsector\tunix_dev\n\n");
    }

    sectors = gclusters * allocblock;

    if (verbose)
	for (i = 0; i < sectors; i++)
	{
	    tr = LTP_TRACK (i);
	    si = LTP_SIDE (i);
	    se = LTP_SCT (i);
	    uxs = tr * gspcyl + gsectors * si + se;
	    if (verbose)
	    {
		printf ("%i\t%i\t%i\t%i\t%i\n", i, tr, si, se, uxs);
	    }
	}
}

#ifdef __WATCOMC__
typedef unsigned short ushort;

ushort biosdisk (ushort service, ushort drive, ushort side, ushort track,
		 ushort sector, ushort nsector, void *buffer)
{
    struct diskinfo_t di;
    ushort status;

    di.drive = drive;
    di.head = side;
    di.track = track;
    di.sector = sector;
    di.nsectors = nsector;
    di.buffer = buffer;
    status = _bios_disk (service, &di);
    return status;
}
#endif

void dump_cluster (int num)
{
    int i, sect;
    unsigned char buf[512];
    int r = 0;

    for (i = 0; i < allocblock; i++)
    {
	short j, k;
	unsigned char *p;
	long fpos;

	sect = num * allocblock + i;
	fpos = LTP (sect);
#ifdef MSDOS
	if (biosdisk (READ, drive, LTP_SIDE (sect), LTP_TRACK (sect), LTP_SECT (sect), 1, buf) & 0xff00)
	    err = r = -1;
#else
	err = lseek (fd, fpos, SEEK_SET);
	if (err < 0)
	    perror ("read_cluster : lseek():");
	r += err = read (fd, buf, gssize);
#endif
	if (err < 0)
	    perror ("read block: read(): ");

	p = buf;
	for (k = 0; k < 32; k++)
	{
	    printf ("%03x : ", k * 16 + fpos);

	    for (j = 0; j < 16; j++)
	    {
		printf (" %02x", *p++);
	    }
	    fputc ('\t', stdout);
	    p -= 16;

	    for (j = 0; j < 16; j++, p++)
	    {
		printf ("%c", isprint (*p) ? *p : '.');
	    }
	    fputc ('\n', stdout);
	}
    }
}

int read_cluster (char *p, int num)
{
    int i, sect;
    int r = 0;

    for (i = 0; i < allocblock; i++)
    {
	sect = num * allocblock + i;

#ifdef MSDOS
	/* biosdisk(RESET, drive , 0, 0, 0, 0, p+gssize*i); */
	if (biosdisk (READ, drive, LTP_SIDE (sect), LTP_TRACK (sect), LTP_SECT (sect), 1, p + gssize * i) & 0xff00)
	    err = r = -1;
#else
	err = lseek (fd, LTP (sect), SEEK_SET);
	if (err < 0)
	    perror ("read_cluster : lseek():");
	r += err = read (fd, p + gssize * i, gssize);
#endif
	if (err < 0)
	    perror ("read block: read(): ");
    }
    return r;
}

void write_cluster (char *p, int num)
{
    int i, sect;
    int r = 0;

    for (i = 0; i < allocblock; i++)
    {
	sect = num * allocblock + i;
#ifdef MSDOS
	if ((r = biosdisk (WRITE, drive, LTP_SIDE (sect), LTP_TRACK (sect),
			   LTP_SECT (sect), 1, p + gssize * i)) & 0xff00)
	    if (r == 0x300)
	    {
		fputs ("Write Protected\n", stderr);
		exit (0);
	    }
	    else
		err = r = -1;
#else
	err = lseek (fd, LTP (sect), SEEK_SET);
	if (err < 0)
	    perror ("write cluster: lseek():");
	r += err = write (fd, p + gssize * i, gssize);
#endif
	if (r < 0)
	    perror ("write cluster: write():");
    }
}

void read_b0fat (int argconv)
{
#ifdef MSDOS
    int status;

    /* reset the disk */

    status = biosdisk (READ, drive, 0, 10, 1, 1, b0);

    /* door change signal ? */

    if (status == 0x06)
	status = biosdisk (READ, drive, 0, 0, 1, 1, b0);

    status = biosdisk (RESET, drive, 0, 0, 0, 0, b0);

    if (status != 0)
	fprintf (stderr, "Disk not ready (continuing...)\n");

    /* read block 0 */

    if ((status = biosdisk (READ, drive, 0, 0, 1, 1, b0)) & 0xff00)
	err = -1;
#else
    err = lseek (fd, SECT0, SEEK_SET);
    if (err < 0)
	perror ("read_cluster0: lseek():");

    err = read (fd, b0, gssize);
#endif
    if (err < 0)
    {
	perror ("read_cluster0: read():");
    }
    /* is this a QL disk ? */

    if ((*(b0) != 'Q') || (*(b0 + 1) != 'L'))
    {
	fprintf (stderr, "\nNot a QL disk !!!\n\n");
	exit (1);
    }
    if ((*(short *) (b0 + 2)) == 0x4135)	/* == QL5A ? */
	ql5a = 1;
    else
	ql5a = 0;

    gtracks = GET_TRACKS ();
    gsectors = GET_SPT ();
    gspcyl = GET_SPCYL ();
    gsides = gspcyl / gsectors;
    goffset = GET_SOFF ();
    bleod = GET_DIRBL ();
    byeod = GET_DIROF ();
    allocblock = GET_SPCL ();
    gclusters = gtracks * gspcyl / allocblock;

    read_ltp ();

    make_convtable (argconv);
    read_fat ();

}

void write_b0fat (void)
{
    int i;
    long upd;

    upd = lwrd (b0 + 0x10);
    upd++;
    wr_lwrd (b0 + 0x10, upd);

    for (i = 0; FAT_FILE (i) == 0xf80; i++)
    {
	write_cluster (b0 + i * allocblock * gssize, i);
    }
}


void read_fat (void)
{
    int i;

    for (i = 0; FAT_FILE (i) == 0xf80; i++)
	read_cluster (b0 + i * allocblock * gssize, i);
}

void read_ltp (void)
{
    int i;

    ltp = (char *) malloc (gspcyl);
    if (ql5a)
	ptl = (char *) malloc (gspcyl);

    for (i = 0; i < gspcyl; i++)
    {
	ltp[i] = *(b0 + 40 + i);
	if (ql5a)
	    ptl[i] = *(b0 + 58 + i);
    }
}


long int match_file (char *fname, char **ent)
{
    int d, i, match, len;
    long int r = 0L;
    char c0, c1;

    len = strlen (fname);

    for (d = 64; d < gssize * bleod + byeod; d += 64)
    {
	match = 1;

	if (wrd (pdir + d + 14) == len)
	{
	    for (i = 0; i < len; i++)
	    {
		c0 = *(pdir + d + 16 + i);
		c0 = tolower (c0);
		c1 = tolower (fname[i]);

		if (c0 != c1)
		{
		    if (!tranql || !(c1 == '.' && c0 == '_'))
		    {
			match = 0;
			break;
		    }
		}
	    }
	    if (match)
	    {
		if(ent)
		{
		    *ent = pdir+d;
		}
		r = d >> 6;
		break;
	    }
	}
    }

    return (r);
}

char *MakeQLName (char *fn, short *n)
{
    char *p, *q;

    if ((p = strrchr (fn, '/')) != NULL)
    {
	p++;
    }
    else if ((p = strrchr (fn, '\\')) != NULL)
    {
	p++;
    }

    if (p == NULL)
	p = fn;
    q = strdup (p);

    *n = min (strlen (q), 36);
    *(q + (*n)) = 0;

    if (tranql)
    {
	for (p = q; *p; p++)
	{
	    if (*p == '.')
		*p = '_';
	}
    }

    return q;
}
#ifdef __linux__
#include <termios.h>
#define BITSON ( ~0 )

void rl_ttyset (int Reset)
{
    static struct termios old;
    struct termios new;

    if (Reset == 0)
    {
	(void) tcgetattr (0, &old);
	new = old;
	new.c_cc[VINTR] = BITSON;
	new.c_cc[VQUIT] = BITSON;
	new.c_lflag &= ~(ECHO | ICANON);
	new.c_iflag &= ~(ISTRIP | INPCK);
	new.c_cc[VMIN] = 1;
	new.c_cc[VTIME] = 0;
	(void) tcsetattr (0, TCSANOW, &new);
    }
    else
	(void) tcsetattr (0, TCSANOW, &old);
}
#endif
static char OWopt = 0;

void writefile (char *fn)
{
    char dati_file[1536];
    int i, filenew;
    int fl = 0, hole;
    long y;
    int offset;
    unsigned char *entry;
    unsigned long free_sect;
    long qdsize = 0;
    int block = 0;
    int end_dati;
    char *qlnam;
    short nlen;
    char *vers = NULL;
    short nvers = -1;
#ifdef __unix__
    struct timeval tv;
    struct timezone tz;
#endif    
    time_t t;

    qlnam = MakeQLName (fn, &nlen);

    if(list)
    {
	fputs(fn, stderr);
	fputc('\n', stderr);
    }
    
    if ((fl = match_file (qlnam, &vers)) != 0)
    {
	if (OWopt != 'A')
	{
	    fprintf (stderr, "file %s exists, overwrite [Y/N/A/Q] : ", qlnam);
#ifdef __unix__
	    rl_ttyset (0);
#endif
	    do
	    {
#ifdef __unix__
		read (0, &OWopt, 1);
#else
# ifdef __MSDOS__
		OWopt = getch ();
# else
		OWopt = 'Q';
# endif
#endif
		OWopt = toupper (OWopt);
	    }
	    while (strchr ("\003ANYQ", OWopt) == NULL);
#ifdef __unix__
	    rl_ttyset (1);
#endif
	    fprintf (stderr, "%c\n", OWopt);
	    if (OWopt == 'N')
	    {
		return;
	    }
	    else
	    {
		if (OWopt == 'Q' || OWopt == 3)
		{
		    exit (2);
		}
	    }
	}
	del_file (fl);
	nvers = wrd(vers + 0x38);
    }

#ifdef MSDOS
    if ((fl = open (fn, O_RDONLY | O_BINARY)) < 0)
    {
	printf ("File %s not open", fn);
	perror (NULL);
	exit (1);
    }
    y = filelength (fl) + 64;	/* Add the 64 byte of QL Header */
#else
    if ((fl = open (fn, O_RDONLY)) == -1)
    {
	perror ("write file: could not open input file ");
	exit (1);
    }
    y = lseek (fl, 0, SEEK_END) + 64;
    if (y < 0)
	perror ("write file: lseek():");
#endif
    {
	long stuff[2];

	lseek (fl, -8, SEEK_END);
	read (fl, stuff, 8);
	if (*stuff == *(long *) "XTcc")
	{
	    qdsize = *(stuff + 1);
	}
    }

    err = lseek (fl, 0, SEEK_SET);	/* reposition to zero!!! */
    if (err < 0)
	perror ("write file: lseek() on input file : ");


    free_sect = GET_FREE ();

    /* controlla se il file  troppo grande */
    if (y > free_sect * gssize)
    {
	printf ("File %s too large (%d %d %d)\n", fn, y, free_sect, gssize);
	exit (1);
    }


    offset = 64;
    while ((lwrd (pdir + offset + 0) + wrd (pdir + offset + 14) > 0)
	   && (offset < gssize * bleod + byeod))
    {
	offset += 64;
    }

    if (offset >= gssize * bleod + byeod)
    {
	hole = 0;
	offset = gssize * bleod + byeod;
/*	printf("## append to directory %d %d \n", bleod, byeod); */
    }
    else
    {
	hole = 1;
/*	printf("## found place in directory\n"); */
    }

    if ((byeod == 0) && ((bleod % allocblock) == 0) && !hole)
    {
	i = alloc_new_cluster (0, block_max);
	if (i < 0)
	{
	    fprintf (stderr, "write file: no free cluster\n");
	    exit (1);
	}
	block++;
	block_dir[block_max] = i;
	block_max++;
/*	printf("## Allocate cluster %d %d\n", bleod, byeod); */
    }

    if (!hole)
	byeod += 64;

    if (byeod == gssize)
    {
	byeod = 0;
	bleod += 1;
    }

    entry = pdir + offset;
    memset(entry, '\0', 64);
    
    wr_lwrd (entry, y);		/* lunghezza del file */
    if (qdsize == 0)
    {
	*(entry + 5) = 0;	/* tipo del file */
	*(long *) (entry + 6) = 0l;
    }
    else
    {
	*(entry + 5) = 1;	/* tipo del file */
	*(long *) (entry + 6) = qdsize;
    }

    wr_wrd (entry + 14, nlen);
    strncpy (pdir + offset + 16, qlnam, nlen);
    free (qlnam);
#ifdef __unix__
    gettimeofday(&tv, &tz);
    t = tv.tv_sec - (60 * tz.tz_minuteswest) + TIME_DIFF;
#else
    t = time(NULL) + TIME_DIFF;
#endif
    wr_lwrd(entry+0x34, t);
    wr_lwrd(entry+0x3c, t);
    nvers++;
    wr_wrd(entry+0x38, nvers);
    
    /* copio il file */

    if (y < 1536)
	end_dati = y;
    else
	end_dati = 1536;
    read (fl, dati_file + 64, end_dati - 64);	/* copio i dati anche nell'header del file */

    memcpy (dati_file, entry, 64);	/* second copy of header */

    /* scrittura del file */

    filenew = (offset + 1) >> 6;
    while (y > 0)
    {
	i = alloc_new_cluster (filenew, block);
	if (i < 0)
	{
	    fprintf (stderr, "filewrite failed : no free cluster\n");
	    exit (1);
	}
	block++;
	write_cluster (dati_file, i);
	y -= allocblock * gssize;

	err = read (fl, dati_file, allocblock * gssize);
	if (err < 0)
	    perror ("write file: read on input file:");
    }
    close (fl);

    SET_FREE (free_sect - block * allocblock);
    /* scrive i byte rimasti liberi; */

    SET_DIRBL (bleod);
    SET_DIROF (byeod);

    write_b0fat ();		/* write_cluster(b0,0); */
    dir_write_back ();
}

int main (int ac, char **av)
{
    int i;

#ifndef __MSDOS__
    int mode = O_RDONLY;

#endif
    int dofmt = 0;
    long np1 = 0, np2 = 0;

    if (ac < 2)
    {
	usage ("too few parameters");
    }

    for (i = 2; i < ac; i++)
    {
	if ((av[i][0] == '-') || (av[i][0] == '/'))
	{
	    switch (av[i][1])
	    {
		case 'f':
		    dofmt = i;
		case 'x':
		case 'r':
		case 'w':
		case 'W':
#ifndef __MSDOS__
		    mode = O_RDWR;
#endif
		    i = ac;
		    break;
	    }
	}
    }

#ifdef MSDOS
    if (av[1][1] != ':')
    {
	fprintf (stderr, "Bad drive %s\n", av[1]);
	usage ("Bad drive: use a: or b:");
    }

    switch (*av[1])
    {
	case 'a':
	case 'A':
	    drive = 0;
	    break;
	case 'b':
	case 'B':
	    drive = 1;
	    break;
	default:
	    usage ("Bad drive: use a: or b:");
	    break;
    }
#else
    fd = open (av[1], mode);
    if (fd < 0)
    {
	perror ("could not open image");
	usage ("image file not opened");
    }
#endif

    if (dofmt)
    {
	format (av[dofmt] + 2, av[dofmt + 1]);
    }

    read_b0fat (0);
    pdir = (unsigned char *) malloc (gssize * allocblock * (bleod + 6));
    read_dir ();

    for (i = 2; i < ac; i++)
    {
	if (av[i][0] == '-')
	{
	    OWopt = 0;
	    switch (av[i][1])
	    {
	        case 'l':
		    list = 1;
		    break;
		case 't':
		    tranql ^= 1;
		    break;
		case 'u':
		    if (av[i][2])
		    {
			np1 = atol (av[i] + 2);
		    }
		    else
		    {
			i++;
			np1 = atol (av[i]);
		    }
		    dump_cluster (np1);
		    break;
		case 'd':
		    np1 = av[i][2] == 'l';
		    print_dir (0, np1);
		    break;
		case 's':
		    print_dir (1, 0);
		    break;
		case 'i':
		    print_info ();
		    break;
		case 'm':
		    print_map ();
		    break;
		case 'c':
		    make_convtable (1);
		    break;
		case 'n':
		    if (av[i][2])
		    {
			np1 = atol (av[i] + 2);
		    }
		    else
		    {
			i++;
			np1 = match_file (av[i], NULL);
		    }
		    if (np1)
		    {
			cat_file (np1);
		    }
		    else
		    {
			fprintf (stderr, "Invalid file\n");
		    }
		    break;
		case 'W':
		    OWopt = 'A';
		case 'w':
		    while (av[i + 1] && *av[i + 1] != '-')
		    {
			i++;
			writefile (av[i]);
		    }
		    break;
		case 'r':
		    if (av[i][2])
		    {
			char *q;

			np1 = strtol (av[i] + 2, &q, 10);
			if (np1 && *q == '-')
			{
			    np2 = atol (q + 1);
			}
			else
			    np2 = np1;
		    }
		    else
		    {
			i++;
			np1 = np2 = match_file (av[i], NULL);
		    }
		    if (np1)
		    {
			do
			{
			    del_file (np1);
			    np1++;
			}
			while (np1 <= np2);
		    }
		    break;

		case 'x':
		    if (av[i][2])
		    {
			np1 = atol (av[i] + 2);
		    }
		    else
		    {
			i++;
			np1 = match_file (av[i], NULL);
		    }
		    if (np1)
		    {
			i++;
			np2 = atol (av[i]);
			if (np2)
			{
			    set_header (np1, np2);
			}
		    }
		    break;
		default:
		    usage ("bad option");
		    break;
	    }
	}
	else
	{
	    np1 = match_file (av[i], NULL);
	    if (np1 == 0)
	    {
		np1 = atol (av[i]);
	    }
	    if (np1)
	    {
		cat_file (np1);
	    }
	}
    }

#ifndef MSDOS
    close (fd);
#endif
    return (0);
}

void read_dir (void)
{
    int i, fn, cl;

    for (i = 0; i < gclusters; i++)
    {

	cl = FAT_CL (i);
	fn = FAT_FILE (i);

	if (fn == 0x00)
	{
	    block_dir[block_max] = i;
	    block_max++;
	    read_cluster (pdir + gssize * allocblock * cl, i);
	}
    }
}

void print_map (void)
{
    int i, fn, cl, fnum = 0, del;
    unsigned char *entry;

    printf ("\nblock\tfile\tpos\n\n");

    for (i = 0; i < gclusters; i++)
    {
	entry = NULL;
	del = 0;
	cl = FAT_CL (i);
	fn = FAT_FILE (i);

	printf ("%d\t%d\t%d\t(%03x, %03x) ", i, fn, cl, fn, cl);

	if ((fn & 0xFF0) == 0xFD0 && (fn & 0xf) != 0xF)
	{
	    printf ("erased %2d\t", fn & 0xF);
	    del = 1;
	    if (cl == 0)
	    {
		read_cluster (b, i);
		entry = b;
		del = 1;
		fnum = fn & 0xf;
	    }
	}

	switch (fn)
	{
	    case 0x000:
		printf ("directory");
		break;
	    case 0xF80:
		printf ("map");
		break;
	    case 0xFDF:
		printf ("unused");
		break;
	    case 0xFEF:
		printf ("bad");
		break;
	    case 0xFFF:
		printf ("not existent");
		break;
	    default:
		if (del == 0)
		{
		    entry = fn * 64 + pdir;
		    del = 0;
		    fnum = fn;
		}
		break;
	}
	if (entry && fnum < gclusters)
	{
	    print_entry (entry, fnum, del, 1);
	}
	printf ("\n");
    }
}

void set_header (int ni, long h)
{
    int i, ii;

    ii = ni * 64;

    if (ii >= bleod * gssize + byeod)
    {
	fprintf (stderr, "filenumber %d out of bounds\n", ni);
	exit (1);
    }
    if (lwrd (pdir + ii) + wrd (pdir + ii + 14) == 0)
    {
	fprintf (stderr, "file %d deleted ??\n", ni);
	exit (1);
    }

    *(pdir + ii + 5) = 1;
    wr_lwrd (pdir + ii + 6, h);

    for (i = 1; i < gclusters; i++)
	if (FAT_FILE (i) == ni && FAT_CL (i) == 0)
	    break;
    /* if (i==gclusters) error .... */

    read_cluster (b, i);
    *(b + 5) = 1;
    wr_lwrd (b + 6, h);
    write_cluster (b, i);
    dir_write_back ();
}

void dir_write_back (void)
{
    int i;

    for (i = 0; i < block_max; i++)
	write_cluster (pdir + gssize * allocblock * i, block_dir[i]);
}


int alloc_new_cluster (int fnum, int iblk)
{
    long fflag, i, ok = 0;

    for (i = lac + 1; i < gclusters; i++)
    {
	fflag = FAT_FILE (i);
	if ((fflag >> 4) == 0xFD)
	{
	    ok = 1;
	    SET_FAT_FILE (i, fnum);
	    SET_FAT_CL (i, iblk);
	    lac = i;
	    break;
	}
    }
    return (ok ? i : -1);
}


void free_cluster (long i)
{

    int fn, dfn;

    if (i > 0)
    {
	fn = FAT_FILE (i);
	dfn = 0xFD0 | (0xf & fn);
	SET_FAT_FILE (i, dfn);
    }
    else
    {
	fprintf (stderr, "freeing cluster 0 ???!!!\n");
	exit (1);
    }
}

void format (char *frmt, char *argfname)
{
    static char ltp_dd[] =
    {
	0, 3, 6, 0x80, 0x83, 0x86, 1, 4,
	7, 0x81, 0x84, 0x87, 2, 5, 8, 0x82, 0x85, 0x88
    };

    static char ptl_dd[] =
    {
	0, 6, 0x0c, 1, 7, 0x0d,
	2, 8, 0x0e, 3, 9, 0x0f, 4, 0x0a, 0x10, 5, 0x0b, 0x11
    };

    static char ltp_hd[] =
    {
	0, 2, 4, 6, 8, 0xa, 0xc, 0xe, 0x10,
	0x80, 0x82, 0x84, 0x86, 0x88, 0x8a, 0x8c, 0x8e, 0x90,
	1, 3, 5, 7, 9, 0xb, 0xd, 0xf, 0x11,
	0x81, 0x83, 0x85, 0x87, 0x89, 0x8b, 0x8d, 0x8f, 0x91
    };

    int cls;
    time_t t;

    t = time (NULL);
    wr_wrd ((b0 + 0xe), t & 0xffff);
    gssize = 512;

    if (*frmt == 'd')		/* 720 K format */
    {

	memcpy (b0, "QL5A          ", 14);
	ql5a = 1;
	strncpy (b0 + 4, argfname, (strlen (argfname) <= 10 ?
				    strlen (argfname) : 10));

	memcpy (b0 + 40, ltp_dd, 18);
	memcpy (b0 + 58, ptl_dd, 18);

	gsides = 2;
	allocblock = 3;
	SET_TRACKS (80);
	gtracks = 80;
	SET_SPT (9);		/* b0+26 */
	gsectors = 9;
	SET_SPCYL (18);		/* b0+28 */
	gspcyl = 18;
	goffset = 5;
	SET_SOFF (5);		/* b0+38 */
	SET_DIRBL (0);		/* b0+34 */
	SET_DIROF (64);		/* b0+36 */
	SET_FREE (1434);	/* b0+20 */

	SET_TOTAL (1440);	/* b0+24 */
	SET_GOOD (1440);	/* b0+22 */
	SET_SPCL (allocblock);	/* b0+32 */

	SET_FAT_FILE (0, 0xF80);	/* FAT entry for FAT */
	SET_FAT_CL (0, 0);
	SET_FAT_FILE (1, 0);	/*  ...  for directory */
	SET_FAT_CL (1, 0);
	gclusters = gtracks * gspcyl / allocblock;
	for (cls = 2; cls < gclusters; cls++)
	{			/* init rest of FAT */
	    SET_FAT_FILE (cls, 0xFDF);
	    SET_FAT_CL (cls, 0xFFF);
	}
	read_ltp ();
	make_convtable (0);
	write_b0fat ();
	close (fd);

	exit (0);
    }
    else if (*frmt == 'h')
    {
	memcpy (b0, "QL5B          ", 14);
	ql5a = 0;

	strncpy (b0 + 4, argfname,
		 (strlen (argfname) <= 10 ? strlen (argfname) : 10));

	memcpy (b0 + 40, ltp_hd, 36);

	gsides = 2;
	allocblock = 3;
	SET_TRACKS (80);
	gtracks = 80;
	SET_SPT (18);		/* b0+26 */
	gsectors = 18;

	SET_SPCYL (36);		/* b0+28 */
	gspcyl = 36;
	goffset = 2;

	SET_SOFF (2);		/* b0+38 */
	SET_DIRBL (0);		/* b0+34 */
	SET_DIROF (64);		/* b0+36 */
	SET_FREE (2871);	/* b0+20 */

	SET_TOTAL (2880);	/* b0+24 */
	SET_GOOD (2880);	/* b0+22 */
	SET_SPCL (allocblock);	/* b0+32 */

	SET_FAT_FILE (0, 0xF80);	/* FAT entry for FAT */
	SET_FAT_CL (0, 0);
	SET_FAT_FILE (1, 0xf80);
	SET_FAT_CL (1, 1);
	SET_FAT_FILE (2, 0);	/*  ...  for directory */
	SET_FAT_CL (2, 0);

	gclusters = gtracks * gspcyl / allocblock;
	for (cls = 3; cls < gclusters; cls++)
	{			/* init rest of FAT */
	    SET_FAT_FILE (cls, 0xFDF);
	    SET_FAT_CL (cls, 0xFFF);
	}
	read_ltp ();
	make_convtable (0);
	write_b0fat ();		/*  write_cluster(b0,0); */
	exit (0);
    }
    exit (1);
}
